/* hash.c
 *
 * hash.c - basic hash table functions
 * By Jon Rifkin <j.rifkin@uconn.edu>
 * Copyright 1999 Jonathan Rifkin
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


/*
------------------------------------------------------------------------
Compile Switches
------------------------------------------------------------------------
*/
#define DEBUG
#undef  DEBUG


/*
------------------------------------------------------------------------
Include files
------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include "hash.h"


/*
------------------------------------------------------------------------
Defines
------------------------------------------------------------------------
*/


/*
------------------------------------------------------------------------
Macros
------------------------------------------------------------------------
*/
#define WRITEMSG \
   printf ("File %s line %d\n", __FILE__, __LINE__); fflush(stdout);
#define WRITETXT(TXT) \
   printf ("File %s line %d: %s\n", __FILE__, __LINE__, TXT); fflush(stdout);
#define WRITEVAR(VAR,FMT) \
   printf ("File %s line %d: ", __FILE__, __LINE__); \
	printf ("%s <", #VAR); \
	printf (#FMT, VAR); \
	printf (">\n"); fflush(stdout);

#define HFREE  free
#define HALLOC calloc
/*
------------------------------------------------------------------------
Typedefs
------------------------------------------------------------------------
*/



/*
------------------------------------------------------------------------
Module variables
------------------------------------------------------------------------
*/
static int nhashbit_m=0;
static int hashmask_m=0;
static int ntable_m=0;
static int AddEnd_m = 0;
static hlist_t *next_m = NULL;
static int      walk_m = 0;
static int      nnode_m = 0;
static int      narray_m = 0;


/*
------------------------------------------------------------------------
Local Function Prototypes
------------------------------------------------------------------------
*/
int makehash(u_char *key, int keylength, int nhashbit)  ;
void hfree (void *p) ;
void *halloc (int n, int size) ;


/*
------------------------------------------------------------------------
Exported Functions
------------------------------------------------------------------------
*/


hlist_t **init_htable (int ntable) {
	int nhashbit = 1;
	ntable_m--;
	while (ntable>>=1) {
		nhashbit++;
	}
	ntable_m = 1 << nhashbit;
	if (ntable_m<=0)
		return NULL;
	nhashbit_m = nhashbit;
	hashmask_m = (1 << nhashbit) - 1;

	return HALLOC(ntable_m, sizeof(hlist_t *));

	}

int finddata (hlist_t **ha, u_char *key, int nkey, u_char **data, int *ndata) {
	int hash;
	hlist_t *t;

	hash = makehash (key, nkey, nhashbit_m);
	if (NULL==ha[hash]) {
		*data = NULL;
		*ndata = 0;
		return 0;
	}

	/*  Search list  */
	t = ha[hash];
	while (NULL!=t) {
		if (t->nkey==nkey && ! memcmp(t->key, key, nkey))  {
			*data  = t->data;
			*ndata = t->ndata;
			return 1;
		}
		t = t->next;
	}
	*data = NULL;
	*ndata = 0;
	return 0;
}


/*  Add node to *front* of list  */
int addnode (hlist_t **ha, u_char *key, int nkey, u_char *data, int ndata) {
	int hash;
	hlist_t **pt;
	hlist_t *t;
	hlist_t *next;

	/*  Find hash  */
	hash = makehash (key, nkey, nhashbit_m);
	
	/*  If table entry is blank, make new entry  */
	if (NULL==ha[hash]) {
		pt = &(ha[hash]);
		narray_m++;

	/*  Search table for existing node  */
	} else {
		pt = &(ha[hash]);
		t  = *pt;
		while (NULL!=t) {

			/*  Existing node with same key, replace the data  */
			if ( t->nkey==nkey &&  ! memcmp(t->key,key,nkey) ) {
				if (t->data)  HFREE(t->data);
				if (NULL==t->data || 0==t->ndata) {
					t->data  = NULL;
					t->ndata = 0;
				} else {
					t->data  = HALLOC(1, ndata);
					t->ndata = ndata;
					memcpy(t->data, data, ndata);	
				}
				return 0;
			}
			pt = &(t->next);
			t  = *pt;
		}
	}

	/*  If reached here then key not found  */
	nnode_m++;

	/*  Make a new node (unattached to list)  */
	t        = HALLOC(1, sizeof(hlist_t));
	t->key   = HALLOC(1, nkey);
	t->data  = HALLOC(1, ndata);
	t->nkey  = nkey;
	t->ndata = ndata;
	t->next  = NULL;
	memcpy(t->key, key, nkey);
	memcpy(t->data, data, ndata);

	/*  Add node to list tail  */
	if (AddEnd_m || NULL==ha[hash])  {
		*pt = t;
	/*  Add node to list head  */
	} else {
		next = ha[hash];
		ha[hash] = t;
		t->next = next;
	}

	return 0;
}


/*  Find first existing node  */
void getfirst (hlist_t **ha) {
	walk_m = 0;
	while (walk_m<ntable_m && NULL==ha[walk_m]) 
		walk_m++;
	next_m = (walk_m<ntable_m) ? ha[walk_m] : NULL;
	}


/*  walk list  */
hlist_t *getnext (hlist_t **ha) {
	/*  Store current node  */
	hlist_t *current = next_m;

	/*  Find next  */
	if ( NULL==next_m )
		return NULL;

	next_m = next_m->next;
	while (NULL==next_m && ++walk_m<ntable_m) {
		next_m = ha[walk_m];
	}

	return current;
}




/*
------------------------------------------------------------------------
Local Functions
------------------------------------------------------------------------
*/
int makehash(u_char *key, int keylength, int nhashbit)  {
	int ikey;
	int tkey;
	int foldkey = 0;
	int ishift = 0;
	int hash;
		
	/*  "fold" original key into four byte key  */
	for (ikey=0; ikey<keylength; ikey++) {
		tkey = (int) key[ikey];
		foldkey ^= (tkey << ishift);
		ishift += 8;
		if (ishift>=32) 
			ishift = 0;
	}

	/*  return hash value from folded key  */
#ifdef DEBUG
	hash = hashmask_m & (foldkey | (foldkey << nhashbit-1) | (foldkey << nhashbit/2));
#endif
   hash = foldkey + (foldkey << 16) - (foldkey >> 16) - 1;
	hash  &= hashmask_m;
	return hash;
}



void hfree (void *p) {
	free(p);
	}


void *halloc (int n, int size) {
	return calloc (n, size);
	}
